#!/usr/bin/python
import os
import logging
from typing import Dict, Optional, List, Any
import json
import yaml

from jinja2 import Environment, FileSystemLoader, Template, BaseLoader
from jinja2.exceptions import TemplateNotFound

from mdg.config import settings
from mdg.tools.filters import get_filters

from mdg.uml import UMLPackage, UMLInstance


logger = logging.getLogger(__name__)


def output_level_copy(source_filename: str, dest_file_template: Template, package: UMLPackage) -> None:
    """ Render a jinja template as pass a UML package as data
    """

    # Render template for UML Package
    dest_filename: str = os.path.abspath(dest_file_template.render(package=package))
    dirname: str = os.path.dirname(dest_filename)

    # make sure computed distination path exists
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    with open(source_filename) as source_fh:
        with open(dest_filename, 'w') as dest_fh:
            dest_fh.write(source_fh.read())

    logger.debug(f"Created {dest_filename}")


def output_level_package(source_template: Template, dest_file_template: Template, package: UMLPackage) -> None:
    """ Render a jinja template as pass a UML package as data
    """

    # Render template for UML Package
    dest_filename: str = os.path.abspath(dest_file_template.render(package=package))
    dirname: str = os.path.dirname(dest_filename)

    # make sure computed distination path exists
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    with open(dest_filename, 'w') as fh:
        fh.write(source_template.render(package=package))

    logger.debug(f"Created {dest_filename}")


def output_level_enum(source_template: Template, dest_file_template: Template, filter_template: Optional[Template], package: UMLPackage) -> None:
    """ Render a jinja template for each enumeration in the supplied package
    """

    # Loop through all enumerations in the UML Package, cheeck the filter result and output if True
    for enum in package.enumerations:
        if filter_template is None or filter_template.render(enum=enum) == "True":
            dest_filename = os.path.abspath(dest_file_template.render(enum=enum))
            dirname = os.path.dirname(dest_filename)

            # make sure computed distination path exists
            if not os.path.exists(dirname):
                os.makedirs(dirname)

            with open(dest_filename, 'w') as fh:
                fh.write(source_template.render(enum=enum))

            logger.debug(f"Created {dest_filename}")


def output_level_class(source_template: Template, dest_file_template: Template, filter_template: Optional[Template], package: UMLPackage) -> None:
    """ Render a jinja template for each class in the supplied package
    """

    # Loop through all classes in the UML Package, cheeck the filter result and output if True
    for cls in package.classes:
        if filter_template is None or filter_template.render(cls=cls) == "True":
            dest_filename = os.path.abspath(dest_file_template.render(cls=cls))
            dirname = os.path.dirname(dest_filename)

            # make sure computed distination path exists
            if not os.path.exists(dirname):
                os.makedirs(dirname)

            with open(dest_filename, 'w') as fh:
                fh.write(source_template.render(cls=cls))

            logger.debug(f"Created {dest_filename}")


def output_level_assoc(source_template: Template, dest_file_template: Template, filter_template: Optional[Template], package: UMLPackage):
    """ Render a jinja template for each association in the supplied package
    """

    # Loop through all associations in the UML Package, cheeck the filter result and output if True
    for assoc in package.associations:
        if filter_template is None or filter_template.render(association=assoc) == "True":
            dest_filename = os.path.abspath(dest_file_template.render(association=assoc))
            dirname = os.path.dirname(dest_filename)

            # make sure computed distination path exists
            if not os.path.exists(dirname):
                os.makedirs(dirname)

            with open(dest_filename, 'w') as fh:
                fh.write(source_template.render(association=assoc))

            logger.debug(f"Created {dest_filename}")


def output_model(package: UMLPackage) -> None:
    """ Loops through model templates in the settings and calls render functions

        Each template consists of:
            dest: The filename of the output
            level: Do we generate a file for each package/class/enumeration/association or root
            source: Path to the jinja2 template
            filter: If supplied, If supplied The template must output "True" for a file to be generated. E.g. "{% if package.classes %}True{% else %}False{% endif %}"
    """
    logger.info("Generating model output for package {}".format(package.path))

    # Create jinja2 filter dict to pass into templates
    filters = get_filters()

    # Create jinja2 environmeent with filters
    import os
    default_templates = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'templates')
    source_env = Environment(loader=FileSystemLoader([settings['templates_folder'], default_templates]))
    source_env.filters = {**source_env.filters, **filters}
    dest_env = Environment(loader=BaseLoader())
    dest_env.filters = {**source_env.filters, **filters}

    # Loop through all template definitions in the config file
    template_definition: Dict
    for template_definition in settings['generation_artifacts']:
        dest_file_template: Template = dest_env.from_string(os.path.join(settings['dest_root'], template_definition['dest']))

        if template_definition['level'] == 'copy':
            if package.parent is None:
                source_file = os.path.join(settings['templates_folder'], template_definition['source'])
                if not os.path.isfile(source_file):
                    source_file = os.path.join(default_templates, template_definition['source'])
                output_level_copy(source_file, dest_file_template, package)
        else:
            # Create jinja2 teemplates for the source file and dest file name
            try:
                source_template: Template = source_env.get_template(template_definition['source'])

                # Filter template is optional and used to skip a file generation.
                filter_template: Optional[Template] = None
                if 'filter' in template_definition.keys():
                    filter_template = Template(template_definition['filter'])

                # Select the output renderer based on the level requested
                if template_definition['level'] == 'package':
                    if filter_template is None or filter_template.render(package=package) == "True":
                        output_level_package(source_template, dest_file_template, package)
                elif template_definition['level'] == 'class':
                    output_level_class(source_template, dest_file_template, filter_template, package)
                elif template_definition['level'] == 'enumeration':
                    output_level_enum(source_template, dest_file_template, filter_template, package)
                elif template_definition['level'] == 'assocication':
                    output_level_assoc(source_template, dest_file_template, filter_template, package)
                elif template_definition['level'] == 'root':
                    if package.parent is None:
                        output_level_package(source_template, dest_file_template, package)
                else:
                    raise ValueError("'{}' is not a valid template level".format(template_definition['level']))
            except TemplateNotFound:
                logger.error(f"Could not find template to render from '{template_definition['source']}' in either the configured templates folder or the default templates folder. Check your templates_folder and source in settings.")

    # Walk through the package hierarchy and recurse output
    child: UMLPackage
    for child in package.children:
        output_model(child)


def output_test_cases(test_cases: List[UMLInstance]) -> None:
    """ Test cases are parse into a list of UML instances. Loop through list and serialise
    """
    if 'test_templates' not in settings or settings['test_templates'] is None:
        logger.warn("No test templates")
        return

    logger.info("Generating test case output")
    filters = get_filters()
    env = Environment(loader=BaseLoader())
    env.filters = {**env.filters, **filters}

    for case in test_cases:
        serialised_json = json.dumps(serialize_instance(case), indent=2)
        serialised_yaml = yaml.dump(serialize_instance(case), Dumper=yaml.CDumper)

        template_definition: Dict
        for template_definition in settings['test_templates']:
            filename_template: Template = env.from_string(template_definition['dest'])
            filename: str = os.path.abspath(filename_template.render(ins=case))
            dirname: str = os.path.dirname(filename)

            # make sure computed distination path exists
            if not os.path.exists(dirname):
                os.makedirs(dirname)

            with open(filename, 'w') as fh:
                if filename[-4:] in ['yaml', '.yml']:
                    fh.write(serialised_yaml)
                else:
                    fh.write(serialised_json)


def serialize_instance(instance: UMLInstance):
    """ Generates a dictionary of attributes, values, dicts and lists from UML instance
        Recurses through associations originaing from supplied instance to serialise sub-objects
    """
    ret: Dict = {}
    filters = get_filters()

    # Extract attributes and their values
    for attr in instance.attributes:
        name = filters['case_attribute'](attr.name)
        ret[name] = attr.value

    # Loop through associations originating from this instance
    for assoc in instance.associations_from:
        dest: Any = assoc.destination
        name = filters['case_attribute'](assoc.destination_name)
        # If the multiplicity is multiple then generate list
        if assoc.destination_multiplicity[1] == '*':
            if name not in ret.keys():
                ret[name] = [serialize_instance(dest), ]
            else:
                ret[name].append(serialize_instance(dest))
        # If multiplicity is singular then generate dict
        else:
            ret[name] = serialize_instance(dest)

    return ret



#!/usr/bin/python
# This file is used as an entry point so requires mdg package to be installed into site-packages
# So after a pip or setup.py install you can just cd to the recipe folder and call mdg_generate

# import sys
import os
import logging


logger = logging.getLogger('mdg')
logger.propagate = False
ch = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s | %(name)s | %(levelname)s | %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)


def generate(args):
    os.environ.setdefault("PYMDG_SETTINGS_MODULE", args.recipe_path)
    from ..generate import generate
    generate()


def validate(args):
    os.environ.setdefault("PYMDG_SETTINGS_MODULE", args.recipe_path)
    from ..uml.validate import validate
    validate()


def dumps(args):
    os.environ.setdefault("PYMDG_SETTINGS_MODULE", args.recipe_path)
    from ..uml import dumps as uml_dumps
    from ..parse import parse

    model_package, test_cases = parse()
    print(uml_dumps(model_package))


def startproject(args):
    """ mdg-tool startproject --source="sqlite:///Customer Model.qea" --parser=sparx --model_package=<GUID> --default_dialect=django config-customer-graphql.yaml
    """
    
    with open(args.project_path, "w") as f:
        f.write(f"source: {args.source}\n")
        f.write(f"parser: {args.parser}\n")
        f.write(f'model_package: "{args.model_package}"\n')
        f.write("dest_root: ./build\n")
        f.write(f"templates_folder: ./templates\n")
        f.write(f"default_dialect: {args.default_dialect}\n")
        f.write("generation_artifacts:\n")


def addtemplate(args):
    """
    """
    templates = {
        "django_model":"""
- dest: "{{package.root_package.name | camelcase}}/{{package.name | camelcase}}/models.py"
  level: package
  source: "django/app/models.py.jinja"
  filter: "{% if package.classes != [] %}True{% else %}False{% endif %}"
""",
        "hasura":"""
- dest: "{{package.root_package.name | camelcase}}/hasura_metadata.json"
  level: root
  source: "hasura.json.jinja"  
"""}
    with open(args.project_path, "a") as f:
        f.write(templates[args.template_type])


def daemon(args):
    os.environ.setdefault("PYMDG_SETTINGS_MODULE", args.recipe_path)
    from mdg.tools.daemon import poller
    poller(args.poll_seconds)


def main():
    import argparse

    parser = argparse.ArgumentParser(description='Model Driven Generation Engine')
    parser.add_argument("-v", "--verbose", help="increase output verbosity", action="store_true")
    subparsers = parser.add_subparsers(title='subcommands', description='valid subcommands', help='subcommand help')

    parser_a = subparsers.add_parser('generate', help='Generate files from a model using a recipe')
    parser_a.add_argument('recipe_path', type=str, help='The path to the recipe config file')
    parser_a.set_defaults(func=generate)

    parser_b = subparsers.add_parser('validate', help='Validate files for a model using a recipe')
    parser_b.add_argument('recipe_path', type=str, help='The path to the recipe config file')
    parser_b.set_defaults(func=validate)

    parser_a = subparsers.add_parser('dumps', help='Outputs parsed model as JSON')
    parser_a.add_argument('recipe_path', type=str, help='The path to the recipe config file')
    parser_a.set_defaults(func=dumps)

    parser_c = subparsers.add_parser('startproject', help='Create project with recipe and templates')
    parser_c.add_argument('-s', '--source', type=str, help="The source model file/DB connection")
    parser_c.add_argument('-p', '--parser', type=str, choices=['sparx', 'sparxdb', 'drawio'], help='The format of the source.')
    parser_c.add_argument('-d', '--default_dialect', type=str, choices=['default', 'django', 'schema', 'java'], help='The dialect to use if not specified per artifact')
    parser_c.add_argument('-m', '--model_package', type=str, help='The ID of the root package in ther source.')
    parser_c.add_argument('project_path', type=str, help='The path to the project')
    parser_c.set_defaults(func=startproject)

    parser_c = subparsers.add_parser('addtemplate', help='Add a generation template to a config')
    parser_c.add_argument('template_type', type=str, choices=["django_model", "hasura"], help='The type of template to add')
    parser_c.add_argument('project_path', type=str, help='The path to the project')
    parser_c.set_defaults(func=addtemplate)

    parser_d = subparsers.add_parser('daemon', help='Poll package versions and run generation jobs on change')
    parser_d.add_argument('recipe_path', type=str, help='The path to the recipe config file')
    parser_d.add_argument('poll_seconds', type=int, help='Seconds between polls')
    parser_d.set_defaults(func=daemon)

    args = parser.parse_args()
    if args.verbose == 0:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.DEBUG)
    try:
        func = args.func
    except AttributeError:  # (https://bugs.python.org/issue16308)
        parser.error("too few arguments")
    func(args)


if __name__ == '__main__':
    main()



#!/usr/bin/python
import logging

from mdg.uml.validate import validate_package
from mdg.generate.render import output_model, output_test_cases
from mdg.parse import parse, ParseError, PARSERS
from mdg.config import settings
from mdg import generation_fields

logger = logging.getLogger('mdg')


def validate_settings():
    logger.debug("Validating settings")
    
    errors = []
    if settings['default_dialect'] not in generation_fields.keys():
        errors.append( f"Settings default_dialect of '{settings['default_dialect']}' is not valid. Generation types are a dictionay of how to map UML attrribute types to physical types. Options are: {generation_fields.keys()}" )

    if settings['parser'] not in PARSERS.keys():
        errors.append( f"Settings parser of '{settings['parser']}' is not valid. Parser must match the input modelling tool type defined in 'source'. Options are: {PARSERS.keys()}" )

    for template_definition in settings['generation_artifacts']:
        levels = ["root","copy","class","package","enumeration","assocication"]
        if template_definition["level"] not in levels:
            errors.append( f"Template level of '{template_definition['level']}' is not valid. Level defines how many times the output will be rendered and the objects passed to the template. Options are: {levels}" )

    if errors != []:
        logger.error(f"Errors in settings: {errors}")
        return False
    
    return True


def generate():
    """ Loads XMI file from settings as an etree
        Calls XMI parser to turn model and tests into python native (see UML metamodel)
        Calls output functions to render for model and tests
    """
    if not validate_settings():
        logger.error("Due to errors in settings, not parsing or generating.")
        return

    logger.debug("generate begin")
    try:
        model_package, test_cases = parse()

        # Validate the parsed model package
        errors = validate_package(model_package)
        if len(errors) > 0:
            print("Validation Errors:")
            for error in errors:
                print("    {}".format(error))

        # TODO: Validate the test cases

        # Generate files from the native python UML
        output_model(model_package)
        if 'test_package' in settings.keys():
            output_test_cases(test_cases)

    except ParseError as e:
        logger.error(e)
    logger.debug("generate end")
